/* Taken from: https://github.com/plone/training/blob/master/_static/custom.css */

.toggle {
    /* background: none repeat scroll 0 0 #e7f2fa; */
    padding: 12px;
    line-height: 24px;
    margin-bottom: 24px;
}

.toggle .admonition-title {
    display: block;
    clear: both;
    cursor: pointer;
}

.toggle .admonition-title:after {
    content: " ▶";
}

.toggle .admonition-title.open:after {
    content: " ▼";
}

.toggle p:last-child {
    margin-bottom: 0;
}
